! Fortran 90 and PVM
! measure of communication performance
! Messages of data exchanged by Ping and Pong: all basic F90 types.
!
! Laurent Gasser (lga@sma.ch)
! Swiss Meteorological Institute
! November 12, 1996

module protocol
use PVM_Constants
use PVM_Interfaces

implicit none

 interface throw_packet
   module procedure throw_integers
   module procedure throw_reals
   module procedure throw_complexes
   module procedure throw_characters
   module procedure throw_logicals
 end interface

 interface catch_packet
   module procedure catch_integers
   module procedure catch_reals
   module procedure catch_complexes
   module procedure catch_characters
   module procedure catch_logicals
 end interface

 public :: throw_packet, catch_packet, specify, await_specifications
 public :: send_bouncing_delays, receive_bouncing_delays
 integer, parameter, public :: PROLOGUE_PACKET_SIZE = 100

private
 integer, parameter :: PACKET_TRANSIT = 200
 integer, parameter :: DELAY_REPORT   = 201
 integer, parameter :: SPECIFICATIONS = 202

contains 

  subroutine specify(to, items)
    implicit none
    integer, intent(in) :: to
    integer, intent(in) :: items(:)
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body specify
    call pvm_initsend(PVMDATADEFAULT, buffer_ID)
    if (buffer_ID < 0) then
       print *," Failed to init the specifications, quit unexpectedly."
       call pvm_perror("Protocol, specify", info)
       call exit(info)
    end if
    call pvm_pack(size(items, 1), info) 
    if (info /= PvmOk) then
       print *, " Failed to pack the number of measures, quit unexpectedly."
       call pvm_perror("Protocol, specify", info)
       call exit(info)
    end if
    call pvm_pack(items, size(items, 1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to pack the packets size, quit unexpectedly."
       call pvm_perror("Protocol, specify", info)
       call exit(info)
    end if
    call pvm_send(to, SPECIFICATIONS, info)
    if (info /= PvmOk) then
       print *, " Failed to send the specifications, quit unexpectedly."
       call pvm_perror("Protocol, specify", info)
       call exit(info)
    end if
  end subroutine specify

  subroutine await_specifications(from, items)
    implicit none
    integer, intent(in) :: from
    integer, pointer    :: items(:)
  ! local declarations
    integer :: buffer_ID   = -1
    integer :: info        = PvmOk
    integer :: nb_measures = -666
!    integer :: bufitems(9) ! uncomment in case of faulty unpack
  ! body await_specifications
    call pvm_recv(from, SPECIFICATIONS, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive the specifications, quit unexpectedly."
       call pvm_perror("Protocol, await_specifications", info)
       call exit(info)
    end if
    call pvm_unpack(nb_measures, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the number of measures, quit unexpectedly."
       call pvm_perror("Protocol, await_specifications", info)
       call exit(info)
    end if

    allocate(items(nb_measures), STAT = info)
    if (info /= 0) then
       print *, " Failed to allocate memory for items, quit unexpectedly."
       call pvm_perror("Protocol, await_specifications", info)
       call exit(info)
    end if
    
    call pvm_unpack(items, size(items, 1), 1, info) ! use next in case of bug
!    call pvm_unpack(bufitems, 9, 1, info) ! uncomment in case of unpack error
    if (info /= PvmOk) then
       print *, " Failed to unpack the largest packet size, quit unexpectedly."
       call pvm_perror("Protocol, await_specifications", info)
       call exit(info)
    end if
!    items(:) = bufitems(:) ! uncomment in case of unpack error
  end subroutine await_specifications

  subroutine send_bouncing_delays(to, delays)
    use metrics
    implicit none
    integer, intent(in)         :: to
    type(durations), intent(in) :: delays
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body send_bouncing_delays
    call pvm_initsend(PVMDATADEFAULT, buffer_ID)
    if (buffer_ID < 0) then
       print *," Failed to init the bouncing delays message, quit unexpectedly."
       call pvm_perror("Protocol, send_bouncing_delays", info)
       call exit(info)
    end if
    call pvm_pack(delays%early, info) 
    if (info /= PvmOk) then
       print *, " Failed to pack the first delay, quit unexpectedly."
       call pvm_perror("Protocol, send_bouncing_delays", info)
       call exit(info)
    end if
    call pvm_pack(delays%late, info) 
    if (info /= PvmOk) then
       print *, " Failed to pack the second delay, quit unexpectedly."
       call pvm_perror("Protocol, send_bouncing_delays", info)
       call exit(info)
    end if
    call pvm_send(to, DELAY_REPORT, info)
    if (info /= PvmOk) then
       print *, " Failed to send the bouncing delays, quit unexpectedly."
       call pvm_perror("Protocol, send_bouncing_delays", info)
       call exit(info)
    end if
  end subroutine send_bouncing_delays

  subroutine receive_bouncing_delays(from, delays)
    use metrics
    implicit none
    integer, intent(in)          :: from
    type(durations), intent(out) :: delays
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body receive_bouncing_delays
    call pvm_recv(from, DELAY_REPORT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive the bouncing delays message, quit unexpectedly."
       call pvm_perror("Protocol, receive_bouncing_delays", info)
       call exit(info)
    end if
    call pvm_unpack(delays%early, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the first delay, quit unexpectedly."
       call pvm_perror("Protocol, receive_bouncing_delays", info)
       call exit(info)
    end if
    call pvm_unpack(delays%late, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the second delay, quit unexpectedly."
       call pvm_perror("Protocol, receive_bouncing_delays", info)
       call exit(info)
    end if
  end subroutine receive_bouncing_delays

  subroutine throw_integers(to, packet, instant)
    use metrics
    implicit none
    integer, intent(in)          :: to
    integer, intent(in)          :: packet(:)
    type(durations), intent(out) :: instant
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_integers
    ! measure time before packing
    call system_clock(count = instant%early)
    ! pack the message
    call pvm_initsend(PVMDATADEFAULT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to init the integers message, quit unexpectedly."
       call pvm_perror("Protocol, throw_integers", info)
       call exit(info)
    end if
    call pvm_pack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to pack integers message, quit unexpectedly."
       call pvm_perror("Protocol, throw_integers", info)
       call exit(info)
    end if
    ! measure time before send
    call system_clock(count = instant%late)
    ! send message
    call pvm_send(to, PACKET_TRANSIT, info)
    if (info /= PvmOk) then
       print *, " Failed to send integers message, quit unexpectedly."
       call pvm_perror("Protocol, throw_integers", info)
       call exit(info)
    end if
  end subroutine throw_integers

   subroutine catch_integers(from, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: from
     integer, intent(out)         :: packet(:)
     type(durations), intent(out) :: instant
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_integers
    ! receive message
    call pvm_recv(from, PACKET_TRANSIT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive integers message, quit unexpectedly."
       call pvm_perror("Protocol, catch_integers", info)
       call exit(info)
    end if
    ! measure time right after receive
    call system_clock(count = instant%early)
    ! unpack the message
    call pvm_unpack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the integers message, quit unexpectedly."
       call pvm_perror("Protocol, catch_integers", info)
       call exit(info)
    end if
    ! measure time after complete unpacking
    call system_clock(count = instant%late)
 
    call pvm_freebuf(buffer_ID, info)
  end subroutine catch_integers

  subroutine throw_reals(to, packet, instant)
    use metrics
    implicit none
    integer, intent(in)          :: to
    real, intent(in)             :: packet(:)
    type(durations), intent(out) :: instant
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_reals
    ! measure time before packing
    call system_clock(count = instant%early)
    ! pack the message
    call pvm_initsend(PVMDATADEFAULT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to init the reals message, quit unexpectedly."
       call pvm_perror("Protocol, throw_reals", info)
       call exit(info)
    end if
    call pvm_pack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to pack reals message, quit unexpectedly."
       call pvm_perror("Protocol, throw_reals", info)
       call exit(info)
    end if
    ! measure time before send
    call system_clock(count = instant%late)
    ! send message
    call pvm_send(to, PACKET_TRANSIT, info)
    if (info /= PvmOk) then
       print *, " Failed to send reals message, quit unexpectedly."
       call pvm_perror("Protocol, throw_reals", info)
       call exit(info)
    end if
   end subroutine throw_reals

   subroutine catch_reals(from, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: from
     real, intent(out)            :: packet(:)
     type(durations), intent(out) :: instant
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_reals
    ! receive message
    call pvm_recv(from, PACKET_TRANSIT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive reals message, quit unexpectedly."
       call pvm_perror("Protocol, catch_reals", info)
       call exit(info)
    end if
    ! measure time right after receive
    call system_clock(count = instant%early)
    ! unpack the message
    call pvm_unpack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the reals message, quit unexpectedly."
       call pvm_perror("Protocol, catch_reals", info)
       call exit(info)
    end if
    ! measure time after complete unpacking
    call system_clock(count = instant%late)

    call pvm_freebuf(buffer_ID, info)
   end subroutine catch_reals

   subroutine throw_complexes(to, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: to
     complex, intent(in)          :: packet(:)
     type(durations), intent(out) :: instant
   ! local declarations
     integer :: buffer_ID = -1
     integer :: info      = PvmOk
   ! body throw_complexes
     ! measure time before packing
     call system_clock(count = instant%early)
     ! pack the message
     call pvm_initsend(PVMDATADEFAULT, buffer_ID)
     if (buffer_ID < 0) then
        print *, " Failed to init the complexes message, quit unexpectedly."
        call pvm_perror("Protocol, throw_complexes", info)
        call exit(info)
     end if
     call pvm_pack(packet, size(packet,1), 1, info) 
     if (info /= PvmOk) then
        print *, " Failed to pack complexes message, quit unexpectedly."
        call pvm_perror("Protocol, throw_complexes", info)
        call exit(info)
     end if
     ! measure time before send
     call system_clock(count = instant%late)
     ! send message
     call pvm_send(to, PACKET_TRANSIT, info)
     if (info /= PvmOk) then
        print *, " Failed to send complexes message, quit unexpectedly."
        call pvm_perror("Protocol, throw_complexes", info)
        call exit(info)
     end if
   end subroutine throw_complexes

   subroutine catch_complexes(from, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: from
     complex, intent(out)         :: packet(:)
     type(durations), intent(out) :: instant
   ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_complexes
    ! receive message
    call pvm_recv(from, PACKET_TRANSIT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive complexes message, quit unexpectedly."
       call pvm_perror("Protocol, catch_complexes", info)
       call exit(info)
    end if
    ! measure time right after receive
    call system_clock(count = instant%early)
    ! unpack the message
    call pvm_unpack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the complexes message, quit unexpectedly."
       call pvm_perror("Protocol, catch_complexes", info)
       call exit(info)
    end if
    ! measure time after complete unpacking
    call system_clock(count = instant%late)

    call pvm_freebuf(buffer_ID, info)
  end subroutine catch_complexes

   subroutine throw_characters(to, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: to
     character, intent(in)        :: packet(:) ! no string, array of char
     type(durations), intent(out) :: instant
   ! local declarations
     integer :: buffer_ID = -1
     integer :: info      = PvmOk
   ! body throw_characters
     ! measure time before packing
     call system_clock(count = instant%early)
     ! pack the message
     call pvm_initsend(PVMDATADEFAULT, buffer_ID)
     if (buffer_ID < 0) then
        print *, " Failed to init the characters message, quit unexpectedly."
        call pvm_perror("Protocol, throw_characters", info)
        call exit(info)
     end if
     call pvm_pack(packet, size(packet,1), 1, info) 
     if (info /= PvmOk) then
        print *, " Failed to pack characters message, quit unexpectedly."
        call pvm_perror("Protocol, throw_characters", info)
        call exit(info)
     end if
     ! measure time before send
     call system_clock(count = instant%late)
     ! send message
     call pvm_send(to, PACKET_TRANSIT, info)
     if (info /= PvmOk) then
        print *, " Failed to send characters message, quit unexpectedly."
        call pvm_perror("Protocol, throw_characters", info)
        call exit(info)
     end if
   end subroutine throw_characters

   subroutine catch_characters(from, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: from
     character, intent(out)       :: packet(:)
     type(durations), intent(out) :: instant
   ! local declarations
     integer :: buffer_ID = -1
     integer :: info      = PvmOk
   ! body throw_characters
     ! receive message
     call pvm_recv(from, PACKET_TRANSIT, buffer_ID)
     if (buffer_ID < 0) then
        print *, " Failed to receive characters message, quit unexpectedly."
        call pvm_perror("Protocol, catch_characters", info)
        call exit(info)
     end if
     ! measure time right after receive
     call system_clock(count = instant%early)
     ! unpack the message
     call pvm_unpack(packet, size(packet,1), 1, info) 
     if (info /= PvmOk) then
        print *, " Failed to unpack the characters message, quit unexpectedly."
        call pvm_perror("Protocol, catch_characters", info)
        call exit(info)
     end if
     ! measure time after complete unpacking
     call system_clock(count = instant%late)

     call pvm_freebuf(buffer_ID, info)
  end subroutine catch_characters

   subroutine throw_logicals(to, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: to
     logical, intent(in)          :: packet(:)
     type(durations), intent(out) :: instant
   ! local declarations
     integer :: buffer_ID = -1
     integer :: info      = PvmOk
   ! body throw_logicals
     ! measure time before packing
     call system_clock(count = instant%early)
     ! pack the message
     call pvm_initsend(PVMDATADEFAULT, buffer_ID)
     if (buffer_ID < 0) then
        print *, " Failed to init the logicals message, quit unexpectedly."
        call pvm_perror("Protocol, throw_logicals", info)
        call exit(info)
     end if
     call pvm_pack(packet, size(packet,1), 1, info) 
     if (info /= PvmOk) then
        print *, " Failed to pack logicals message, quit unexpectedly."
        call pvm_perror("Protocol, throw_logicals", info)
        call exit(info)
     end if
     ! measure time before send
     call system_clock(count = instant%late)
     ! send message
     call pvm_send(to, PACKET_TRANSIT, info)
     if (info /= PvmOk) then
        print *, " Failed to send logicals message, quit unexpectedly."
        call pvm_perror("Protocol, throw_logicals", info)
        call exit(info)
     end if
  end subroutine throw_logicals
   
   subroutine catch_logicals(from, packet, instant)
     use metrics
     implicit none
     integer, intent(in)          :: from
     logical, intent(out)         :: packet(:)
     type(durations), intent(out) :: instant
  ! local declarations
    integer :: buffer_ID = -1
    integer :: info      = PvmOk
  ! body throw_logicals
    ! receive message
    call pvm_recv(from, PACKET_TRANSIT, buffer_ID)
    if (buffer_ID < 0) then
       print *, " Failed to receive logicals message, quit unexpectedly."
       call pvm_perror("Protocol, catch_logicals", info)
       call exit(info)
    end if
    ! measure time right after receive
    call system_clock(count = instant%early)
    ! unpack the message
    call pvm_unpack(packet, size(packet,1), 1, info) 
    if (info /= PvmOk) then
       print *, " Failed to unpack the logicals message, quit unexpectedly."
       call pvm_perror("Protocol, catch_logicals", info)
       call exit(info)
    end if
    ! measure time after complete unpacking
    call system_clock(count = instant%late)

    call pvm_freebuf(buffer_ID, info)
   end subroutine catch_logicals

end module protocol
